/*
** RasterLayout
**
**	This module must build up the Display version of the
**	raster image, creating the XImage structure. This should be
**	called whenever the content, image dimensions or anything
**	affecting the presentation changes.
**
** Copyright 1992, 1993 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author may be contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/

#if SYSV_INCLUDES
#	include <memory.h>
#	include <malloc.h>
#endif
#if ANSI_INCLUDES
#	include <stddef.h>
#	include <stdlib.h>
#endif
#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xew/XewInit.h>
#include <X11/Xew/RasterP.h>

#include "RasterLayout.h"
#include "RasterColor.h"
#include "Raster24to8.h"

#define MONO(rd,gn,bl) (((rd)*11 + (gn)*16 + (bl)*5) >> 5)  /*.33R+ .5G+ .17B*/


/*
** Unsupported
**	Generate XtAppWarningMsg about yet unsupported Image/Display
**	combination.
*/
static void Unsupported(w)
XeRasterWidget w;
    {
	int pixelsize;
	XImage *image = w->raster.image;
	char buf[200];
	static char fmt[] = "%s image on %s visual:Image pixel=%d display pixel=%d. Not supported";

	pixelsize = w->raster.bits_per_component * w->raster.samples_per_pixel;
	sprintf(buf, fmt,
		w->raster.class == XeRasterClass_BILEVEL ? "Bi-Level" :
		w->raster.class == XeRasterClass_GRAYSCALE ? "Grayscale" :
		w->raster.class == XeRasterClass_PALETTE ? "Palette" :
		w->raster.class == XeRasterClass_FULLCOLOR ? "Full color" :
		"Unknown class",
		w->raster.visual->class == TrueColor ? "TrueColor" :
		w->raster.visual->class == PseudoColor ? "PseudoColor" :
		w->raster.visual->class == DirectColor ? "DirectColor" :
		w->raster.visual->class == GrayScale ? "GrayScale" :
		w->raster.visual->class == StaticColor ? "StaticColor" :
		w->raster.visual->class == StaticGray ? "StaticGray" :
		"Unknown visual",
		pixelsize, (int)image->bits_per_pixel);
	XeWidgetWarningMsg
		((Widget)w, "rasterLayoutWarning", buf, (String *)NULL, 0);
    }

/*
** Functions InitSpline, EvalSpline and GenerateFSGamma have been taken
** from John Bradley's xv package. Currently these are only used in
** computing the Floyd-Steinberg gamma curve. Should probably just compute
** the curve offline and compile it statically in... --msa
** The following copyright applies these three functions:
*/
/*
 * Copyright 1989, 1990, 1991, 1992 by John Bradley and
 *                       The University of Pennsylvania
 *
 * Permission to use, copy, and distribute for non-commercial purposes,
 * is hereby granted without fee, providing that the above copyright
 * notice appear in all copies and that both the copyright notice and this
 * permission notice appear in supporting documentation. 
 *
 * The software may be modified for your own purposes, but modified versions
 * may not be distributed.
 *
 * This software is provided "as is" without any expressed or implied warranty.
 *
 * The author may be contacted via:
 *    US Mail:   John Bradley
 *               GRASP Lab, Room 301C
 *               3401 Walnut St.  
 *               Philadelphia, PA  19104
 *
 *    Phone:     (215) 898-8813
 *    EMail:     bradley@cis.upenn.edu       
 */
static void InitSpline(x, y, n, y2)
int *x, *y;
int n;
double *y2;
    {
	/* given arrays of data points x[0..n-1] and y[0..n-1], computes the
	   values of the second derivative at each of the data points
	   y2[0..n-1] for use in the splint function */
	
	int i,k;
	double p,qn,sig,un,u[10];
	
	y2[0] = u[0] = 0.0;
	
	for (i=1; i<n-1; i++) 
	    {
		sig = ((double) x[i]-x[i-1]) / ((double) x[i+1] - x[i-1]);
		p = sig * y2[i-1] + 2.0;
		y2[i] = (sig-1.0) / p;
		u[i] = (((double) y[i+1]-y[i]) / (x[i+1]-x[i])) - 
			(((double) y[i]-y[i-1]) / (x[i]-x[i-1]));
		u[i] = (6.0 * u[i]/(x[i+1]-x[i-1]) - sig*u[i-1]) / p;
	    }
	qn = un = 0.0;
	
	y2[n-1] = (un-qn*u[n-2]) / (qn*y2[n-2]+1.0);
	for (k=n-2; k>=0; k--)
		y2[k] = y2[k]*y2[k+1]+u[k];
    }

static double EvalSpline(xa, ya, y2a, n, x)
int xa[], ya[];
double y2a[];
int n;
double x;
    {
	int klo,khi,k;
	double h,b,a;
	
	klo = 0;
	khi = n-1;
	while (khi-klo > 1)
	    {
		k = (khi+klo) >> 1;
		if (xa[k] > x) khi = k;
		else klo = k;
	    }
	h = xa[khi] - xa[klo];
	if (h==0.0)
	    {
		/* ...add some proper error handling here later!!! --msa */
		printf("bad xvalues in splint\n");
		exit(1);
	    }
	a = (xa[khi]-x)/h;
	b = (x-xa[klo])/h;
	return (a*ya[klo] + b*ya[khi] +
		((a*a*a-a)*y2a[klo] + (b*b*b-b)*y2a[khi]) * (h*h) / 6.0);
    }

static unsigned char fsgamcr[256];
static int fsgamcr_initialized = 0;

static void GenerateFSGamma()
    {
	/* this function generates the Floyd-Steinberg gamma curve (fsgamcr)

	   This function generates a 4 point spline curve to be used as a 
	   non-linear grey 'colormap'.  Two of the points are nailed down at
	   0,0 and 255,255, and can't be changed.  You specify the other two.
	   If you specify points on the line (0,0 - 255,255), you'll get the
	   normal linear reponse curve.  If you specify points of 50,0 and
	   200,255, you'll get grey values of 0-50 to map to black (0), and
	   grey values of 200-255 to map to white (255) (roughly).  Values
	   between 50 and 200 will cover the output range 0-255.  The reponse
	   curve will be slightly 's' shaped. */
	
	int i,j;
	static int x[4] = {0,32,224,255};
	static int y[4] = {0, 0,255,255};
	double yf[4];
	
	InitSpline(x, y, 4, yf);
	
	for (i=0; i<256; i++)
	    {
		j = (int) EvalSpline(x, y, yf, 4, (double) i);
		if (j<0)
			j=0;
		else if (j>255)
			j=255;
		fsgamcr[i] = j;
	    }
    }


/*
** ************************
** End Of Extract from 'xv'
** ************************
*/

/*
** Palette8to1
**	Scale/dither 8 bit palette image into 1 bit image using
**	Floyd-Steinberg dithering:
**
**		----   x    7/16
**		3/16  5/16  1/16
**
*/
static void Palette8to1
	(map, dst, dstW, dstH, dstline, src, srcW, srcH, srcline)
unsigned char *map;
unsigned char *dst; int dstW, dstH, dstline;
unsigned char *src; int srcW, srcH, srcline;
    {
	static unsigned char mask[] = 
	    {
		0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1,
	    };
	register int sCol, dCol;
	int sRow, dRow;
	int Pweight, Psum, Qsum, count;
	int err, *nxt_err, *lst_err, *tmp, fwd_err;
	int *errbuf = (int *)malloc(2 * (dstW+2) * sizeof(int));
	
	nxt_err = errbuf + 1;
	lst_err = nxt_err + dstW + 1;
	memset((char *)lst_err, 0, dstW * sizeof(int));
	memset((void *)dst, 0, dstH * dstline);
	Qsum = 0;
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		Qsum += dstH;
		if (Qsum < srcH)
			continue;
		do
		    {
			Psum = Pweight = count = 0;
			nxt_err[0] = 0;
			fwd_err = 0;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				Pweight += map[src[sCol]];
				Psum += dstW;
				count += 1;
				if (Psum < srcW)
					continue;
				Pweight = (Pweight + count / 2) / count;
				do
				    {
					err = Pweight+fwd_err+lst_err[dCol];
					if (err < 128)
						dst[dCol/8] |= mask[dCol%8];
					else
						err -= 255;
					fwd_err = (err*7)/16;
					nxt_err[dCol-1] += (err*3)/16;
					nxt_err[dCol] += (err*5)/16;
					nxt_err[dCol+1] = err/16;
					dCol += 1;
				    }
				while ((Psum -= srcW) > srcW && dCol < dstW);
				count = Pweight = 0;
			    }
			tmp = nxt_err;
			nxt_err = lst_err;
			lst_err = tmp;
			dst += dstline;
			dRow += 1;
			
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
	    }
	free((void *)errbuf);
    }

static void Palette4to1
	(map, dst, dstW, dstH, dstline, src, srcW, srcH, srcline)
unsigned char *map;
unsigned char *dst; int dstW, dstH, dstline;
unsigned char *src; int srcW, srcH, srcline;
    {
    }

/*
** RGB24to1
**	Scale/dither 24 bit RGB image into 1 bit image using
**	Floyd-Steinberg dithering:
**
**		----   x    7/16
**		3/16  5/16  1/16
**
*/
static void RGB24to1(samples, dst,dstW,dstH,dstline, src,srcW,srcH,srcline)
int samples;
unsigned char *dst; int dstW, dstH, dstline;
unsigned char *src; int srcW, srcH, srcline;
    {
	static unsigned char mask[] = 
	    {
		0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1,
	    };
	register int sCol, dCol;
	int sRow, dRow;
	int Pweight, Psum, Qsum, count;
	int err, *nxt_err, *lst_err, *tmp, fwd_err;
	int *errbuf = (int *)malloc(2 * (dstW+2) * sizeof(int));
	unsigned char *s;

	nxt_err = errbuf + 1;
	lst_err = nxt_err + dstW + 1;
	memset((void *)lst_err, 0, dstW * sizeof(int));
	memset((void *)dst, 0, dstH * dstline);
	Qsum = 0;
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		Qsum += dstH;
		if (Qsum < srcH)
			continue;
		do
		    {
			Psum = Pweight = count = 0;
			nxt_err[0] = 0;
			fwd_err = 0;
			s = src;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				Pweight += fsgamcr[MONO(s[0], s[1], s[2])];
				s += samples;
				Psum += dstW;
				count += 1;
				if (Psum < srcW)
					continue;
				Pweight = (Pweight + count / 2) / count;
				do
				    {
					err = Pweight+fwd_err+lst_err[dCol];
					if (err < 128)
						dst[dCol/8] |= mask[dCol%8];
					else
						err -= 255;
					fwd_err = (err*7)/16;
					nxt_err[dCol-1] += (err*3)/16;
					nxt_err[dCol] += (err*5)/16;
					nxt_err[dCol+1] = err/16;
					dCol += 1;
				    }
				while ((Psum -= srcW) > srcW && dCol < dstW);
				count = Pweight = 0;
			    }
			tmp = nxt_err;
			nxt_err = lst_err;
			lst_err = tmp;
			dst += dstline;
			dRow += 1;
			
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
	    }
	free((void *)errbuf);
    }

/*
** ImageScaleNtoM
**	Scale N*8 bit image into M*8 bit image dropping or replicating
**	pixels. (M <= N required).
*/
static void ImageScaleNtoM(M,dst,dstW,dstH,dstline, N,src,srcW,srcH,srcline)
int M; unsigned char *dst; int dstW, dstH, dstline;
int N; unsigned char *src; int srcW, srcH, srcline;
    {
	register int sCol, dCol;
	int sRow, dRow;
	int Psum, Qsum;
	unsigned char *s, *d;

	Qsum = 0;
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		Qsum += dstH;
		if (Qsum < srcH)
			continue;
		do
		    {
			s = src;
			d = dst;
			Psum = 0;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol, s += N)
			    {
				Psum += dstW;
				if (Psum < srcW)
					continue;
				do
				    {
					register int i = 0;
					
					for ( ;i < M; i++)
						d[i] = s[i];
					dCol++;
					d += M;
				    }
				while ((Psum -= srcW) > srcW && dCol < dstW);
			    }
			dst += dstline;
			dRow += 1;
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
	    }
    }

/*
** ImageScale4to8
**	Scale 4 bit image into 8 bit image dropping or replicating pixels
*/
static void ImageScale4to8(dst, dstW, dstH, dstline, src, srcW, srcH, srcline)
unsigned char *dst; int dstW, dstH, dstline;
unsigned char *src; int srcW, srcH, srcline;
    {
	register int sCol, dCol;
	int sRow, dRow;
	int Pweight, Psum, Qsum;

	Qsum = 0;
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		Qsum += dstH;
		if (Qsum < srcH)
			continue;
		do
		    {
			Psum = 0;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				Psum += dstW;
				if (Psum < srcW)
					continue;
				Pweight = src[sCol>>1];
				if ((sCol & 1) == 0)
					Pweight >>= 4;
				Pweight &= 0x0f;
				do
					dst[dCol++] = Pweight;
				while ((Psum -= srcW) > srcW && dCol < dstW);
			    }
			dst += dstline;
			dRow += 1;
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
	    }
    }

/*
** ImageScale1to1
**	Scale 1 bit image into 1 bit image and apply 
**	Floyd-Steinberg dithering while doing it...
**
**		----   x    7/16
**		3/16  5/16  1/16
**
*/
static void ImageScale1to1(dst, dstW, dstH, dstline, src, srcW, srcH, srcline)
unsigned char *dst; int dstW, dstH, dstline;
unsigned char *src; int srcW, srcH, srcline;
    {
	static unsigned char mask[] = 
	    {
		0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1,
	    };
	register int sCol, dCol;
	int sRow, dRow;
	int Psum, Qsum, wCount, hCount = 0, val;
	int *Pweight = (int *)malloc(dstW * sizeof(int));
	int err, *nxt_err, *lst_err, *tmp, fwd_err;
	int *errbuf = (int *)malloc(2 * (dstW+2) * sizeof(int));

	nxt_err = errbuf + 1;
	lst_err = nxt_err + dstW + 1;
	memset((void *)lst_err, 0, dstW * sizeof(int));
	memset((void *)Pweight, 0, dstW * sizeof(int));
	memset((void *)dst, 0, dstH * dstline);
	Qsum = 0;
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		Qsum += dstH;
		hCount += 1;
		if (Qsum < srcH)
		    {
			Psum  = 0;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				Pweight[dCol] +=
					(src[sCol/8] & mask[sCol%8]) != 0;
				Psum += dstW;
				if (Psum < srcW)
					continue;
				do
					dCol += 1;
				while ((Psum -= srcW) > srcW && dCol < dstW);
			    }
			continue;
		    }
		
		do
		    {
			Psum = wCount = 0;
			nxt_err[0] = 0;
			fwd_err = 0;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				Pweight[dCol] +=
					(src[sCol/8] & mask[sCol%8]) != 0;
				Psum += dstW;
				wCount += 1;
				if (Psum < srcW)
					continue;
				wCount *= hCount;
				val = (255*Pweight[dCol] + wCount/2) / wCount;
				val = fsgamcr[val];
				do
				    {
					err = val+fwd_err+lst_err[dCol];
					if (err > 127)
					    {
						dst[dCol/8] |= mask[dCol%8];
						err -= 255;
					    }
					fwd_err = (err*7)/16;
					nxt_err[dCol-1] += (err*3)/16;
					nxt_err[dCol] += (err*5)/16;
					nxt_err[dCol+1] = err/16;
					Pweight[dCol] = 0;
					dCol += 1;
				    }
				while ((Psum -= srcW) > srcW && dCol < dstW);
				wCount = 0;
			    }
			tmp = nxt_err;
			nxt_err = lst_err;
			lst_err = tmp;
			dst += dstline;
			dRow += 1;
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
		hCount = 0;
	    }
	free((void *)Pweight);
	free((void *)errbuf);
    }


/*
** ImageScale1to8
**	Scale 1 bit image into 8 bit grayscale image
*/
static void ImageScale1to8(dst, dstW, dstH, dstline, src, srcW, srcH, srcline)
unsigned char *dst; int dstW, dstH, dstline;
unsigned char *src; int srcW, srcH, srcline;
    {
	static unsigned char mask[] = 
	    {
		0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1,
	    };
	register int sCol, dCol;
	int sRow, dRow;
	int Pweight, Psum, Qsum, count;

	memset((void *)dst, 0, dstH * dstline);
	Qsum = 0;
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		Qsum += dstH;
		if (Qsum < srcH)
			continue;
		do
		    {
			Psum = Pweight= count = 0;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				Pweight += (src[sCol/8] & mask[sCol%8]) != 0;
				count += 1;
				Psum += dstW;
				if (Psum < srcW)
					continue;
				Pweight = (Pweight * 255 + count/2) / count;
				do
				    {
					dst[dCol] = Pweight;
					dCol += 1;
				    }
				while ((Psum -= srcW) > srcW && dCol < dstW);
				Pweight = count = 0;
			    }
			dst += dstline;
			dRow += 1;
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
	    }
    }


/*
** ImageToTrueColor
**	Convert and scale a subclass of color images into TrueColor
**	images (XImage format).
**	This is very generic and *slow* function. This function should not
**	be used for TrueColor that is wide enough to hold all color
**	components in full (that is, each >=8 bits).
**
**	The input colour values are assumed to be 8 bits for each
**	component (red, green, blue). The method of accessing the
**	input colours is determined by the number of samples as
**	follows:
**
**	M = 1,	The RGB values are fetched from the r,g,b colour
**		color tables using the sample as index (palette
**		colour images).
**
**	M = 2,	Illegal value. (not checked)
**
**	M > 2,	The first three bytes of the sample are taken as
**		RGB values.
*/
static int highbit(ul)
unsigned long ul;
    {
	/* returns position of highest set bit in 'ul' as an integer (0-31),
		   or -1 if none */
	
		int i;
	for (i=31; ((ul&0x80000000) == 0) && i>=0;  i--, ul<<=1);
	return i;
    }

static void ImageToTrueColor
	(M, N, r,g,b,rm,gm,bm,dst,dstW,dstH,dstline,src,srcW,srcH,srcline)
int M;				/* Number of bytes in sample */
int N;				/* Number of bytes in TrueColor Pixel */
unsigned char *r, *g, *b;	/* RGB 8bit Colour Mapping */
unsigned long rm, gm, bm;	/* red/green/blue masks */
unsigned char *dst; int dstW, dstH, dstline;
	unsigned char *src; int srcW, srcH, srcline;
    {
	register int sCol, dCol;
	int sRow, dRow;
	int Psum, Qsum, hCount = 0, wCount;
	unsigned long *rW =
		(unsigned long *)XtMalloc(3 * sizeof(long) * (dstW+1));
	unsigned long *gW = rW + (dstW+1);
	unsigned long *bW = gW + (dstW+1);
	unsigned long rem, gem, bem;
	int rs, gs, bs;

	rs = 7 - highbit(rm);
	gs = 7 - highbit(gm);
	bs = 7 - highbit(bm);

	rm = rs > 0 ? (rm << rs) : (rm >> -rs);
	gm = gs > 0 ? (gm << gs) : (gm >> -gs);
	bm = bs > 0 ? (bm << bs) : (bm >> -bs);
	rem = 0xFF & ~rm;
	gem = 0xFF & ~gm;
	bem = 0xFF & ~bm;

	Qsum = 0;
	memset((char *)rW, 0, 3 * sizeof(long) * (dstW+1));
	for (dRow = 0, sRow = 0; dRow < dstH; src += srcline, ++sRow)
	    {
		unsigned char *s;

		Qsum += dstH;
		hCount += 1;
		if (Qsum < srcH)
		    {
			Psum  = 0;
			s = src;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				if (M < 3)
				    {
					int pix = s[0];

					rW[dCol] += r[pix];
					gW[dCol] += g[pix];
					bW[dCol] += b[pix];
					Psum += dstW;
				    }
				else
				    {
					rW[dCol] += s[0];
					gW[dCol] += s[1];
					bW[dCol] += s[2];
				    }
				s += M;
				if (Psum < srcW)
					continue;
				do
					dCol += 1;
				while ((Psum -= srcW) > srcW && dCol < dstW);
			    }
			continue;
		    }
		do
		    {
			long re = 0, ge = 0, be = 0;

			Psum = wCount = 0;
			s = src;
			for (dCol = 0, sCol = 0; dCol < dstW; ++sCol)
			    {
				unsigned long rv, gv, bv, pixel;

				if (M < 3)
				    {
					int pix = src[sCol];
				
					rW[dCol] += r[pix];
					gW[dCol] += g[pix];
					bW[dCol] += b[pix];
				    }
				else
				    {
					rW[dCol] += s[0];
					gW[dCol] += s[1];
					bW[dCol] += s[2];
				    }
				s += M;
				wCount += 1;
				Psum += dstW;
				if (Psum < srcW)
					continue;
				wCount *= hCount;
				rv = (rW[dCol] + wCount/2) / wCount;
				gv = (gW[dCol] + wCount/2) / wCount;
				bv = (bW[dCol] + wCount/2) / wCount;
				if ((re += rv) < 0)
					rv = 0;
				else
				    {
					rv = (re > rm ? ~0 : re) & rm;
					re -= rv | rem;
					rv = (rs<0) ? (rv<<-rs):(rv>>rs);
				    }
				if ((ge += gv) < 0)
					gv = 0;
				else
				    {
					gv = (ge > gm ? ~0 : ge) & gm;
					ge -= gv | gem;
					gv = (gs<0) ? (gv<<-gs):(gv>>gs);
				    }
				if ((be += bv) < 0)
					bv = 0;
				else
				    {
					bv = (be > bm ? ~0 : be) & bm;
					be -= bv | bem;
					bv = (bs<0) ? (bv<<-bs):(bv>>bs);
				    }
				pixel = rv  | gv | bv;
				do
				    {
					/* Following is horribly unportable,
					   assumes short=2 and long=4 bytes
					   --msa */
					if (N == 1)
						dst[dCol] = pixel;
					else if (N == 2)
						*((short *)dst + dCol) = pixel;
					else
						*((long *)dst + dCol) = pixel;
					dCol += 1;
				    }
				while ((Psum -= srcW) > srcW && dCol < dstW);
				wCount = 0;
			    }
			dst += dstline;
			dRow += 1;
		    }
		while ((Qsum -= srcH) > srcH && dRow < dstH);
		hCount = 0;
		memset((char *)rW, 0, 3 * sizeof(long) * (dstW+1));
	    }
	XtFree((void *)rW);
    }

static void LayoutBiLevelImage (w, width, height)
XeRasterWidget w;
Dimension width, height;
    {
	XImage *image = w->raster.image;

	if (image->format == XYBitmap)
	    {
		ImageScale1to1(image->data, width, height,
			       image->bytes_per_line,
			       w->raster.data, w->raster.width,
			       w->raster.height, w->raster.bytes_per_line);
		return;
	    }
	else switch(w->raster.visual->class)
	    {
	    case TrueColor:
		break;
	    case GrayScale:
	    case StaticGray:
	    case DirectColor:
	    case PseudoColor:
		if (image->bits_per_pixel == 8)
		    {
			unsigned char *r, *g, *b;
			int save_bps, i, x, v;
			
			ImageScale1to8(image->data, width, height,
				       image->bytes_per_line,
				       w->raster.data, w->raster.width,
				       w->raster.height,
				       w->raster.bytes_per_line);
			/*
			** Create a dummy colormap (check colour
			** correctness later!!!)
			*/
			XtFree((char *)w->raster.color_map);
			w->raster.color_map =
				(unsigned char *)XtMalloc(3 * 256);
			r = w->raster.color_map;
			g = r + 256;
			b = g + 256;
			x = 256 >> 8;
			v = 255;
			for (i = 0 ; i < 256; i++, v -= x)
				r[i] = g[i] = b[i] = v;
			save_bps = w->raster.bits_per_sample;
			w->raster.bits_per_sample = 8;
			XeRasterAllocColors(w);
			w->raster.bits_per_sample = save_bps;
			XtFree((char *)w->raster.color_map);
			w->raster.color_map = NULL;
			return;
		    }
		break;
	    case StaticColor:
	    default:
		break;
	    }
	Unsupported(w);
    }

static void LayoutGrayScaleImage(w, width, height)
XeRasterWidget w;
Dimension width, height;
    {
	/* Native grayscale not yet implemented. Not so serious omission,
	   as currently all grayscale images will appear as pseudocolor
	   images with colormap computed --msa */
	Unsupported(w);
    }

static void LayoutPaletteColorImage(w, width, height)
XeRasterWidget w;
Dimension width, height;
    {
	int pixelsize;
	XImage *image = w->raster.image;

	pixelsize = w->raster.bits_per_component * w->raster.samples_per_pixel;
	if (image->format == XYBitmap)
	    {
		unsigned char graymap[256], *r, *g, *b;
		int i;
		int mapsize = 1 << w->raster.bits_per_sample;

		r = w->raster.color_map;
		g = r + mapsize;
		b = g + mapsize;
		if (mapsize > sizeof(graymap))
			mapsize = 256;
		for (i = 0; i < mapsize; i++)
			graymap[i] = fsgamcr[MONO(r[i],g[i],b[i]) & 0xFF];
		if (pixelsize == 8)
			Palette8to1(graymap, image->data, width, height,
				    image->bytes_per_line,
				    w->raster.data, w->raster.width,
				    w->raster.height,
				    w->raster.bytes_per_line);
		else if (pixelsize == 4)
			Palette4to1(graymap, image->data, width, height,
				    image->bytes_per_line,
				    w->raster.data, w->raster.width,
				    w->raster.height,
				    w->raster.bytes_per_line);
		else
			Unsupported(w);
		return;
	    }
	else switch (w->raster.visual->class)
	    {
	    case StaticColor:
	    case StaticGray:
	    case DirectColor:
	    case PseudoColor:
		if (image->bits_per_pixel == 8)
		    {
			if (pixelsize == 8)
				ImageScaleNtoM(1,
					       image->data, width, height,
					       image->bytes_per_line,
					       1,
					       w->raster.data, w->raster.width,
					       w->raster.height,
					       w->raster.bytes_per_line);
			else if (pixelsize == 4)
				ImageScale4to8(image->data, width, height,
					       image->bytes_per_line,
					       w->raster.data, w->raster.width,
					       w->raster.height,
					       w->raster.bytes_per_line);
			else
				break; /* no support! */
			XeRasterAllocColors(w);
			return;
		    }
		break;
	    case TrueColor:
		if (pixelsize == 8)
		    {
			ImageToTrueColor
				(1, image->bits_per_pixel / 8,
				 w->raster.color_map,
				 w->raster.color_map
				   + (1<<w->raster.bits_per_sample),
				 w->raster.color_map
				   + (2<<w->raster.bits_per_sample),
				 image->red_mask,
				 image->green_mask,
				 image->blue_mask,
				 image->data, width, height,
				 image->bytes_per_line,
				 w->raster.data, w->raster.width,
				 w->raster.height,
				 w->raster.bytes_per_line);
			return;
		    }
		break;
	    case GrayScale:
	    default:
		break;
	    }
	Unsupported(w);
    }

static void LayoutFullColorImage(w, width, height)
XeRasterWidget w;
Dimension width, height;
    {
	int pixelsize;
	XImage *image = w->raster.image;

	pixelsize = w->raster.bits_per_component * w->raster.samples_per_pixel;
	if (image->format == XYBitmap)
	    {
		if (w->raster.bits_per_sample == 8)
			RGB24to1(w->raster.samples_per_pixel,
				 image->data, width, height,
				 image->bytes_per_line,
				 w->raster.data, w->raster.width,
				 w->raster.height,
				 w->raster.bytes_per_line);
		else
			Unsupported(w);
		return;

	    }
	else switch(w->raster.visual->class)
	    {
	    case TrueColor:
		if (w->raster.bits_per_component == 8 &&
		    w->raster.samples_per_pixel > 2)
		    {
			ImageToTrueColor
				(w->raster.samples_per_pixel,
				 image->bits_per_pixel / 8,
				 (unsigned char *)NULL, /* no r table */
				 (unsigned char *)NULL, /* no g table */
				 (unsigned char *)NULL, /* no b table */
				 image->red_mask,
				 image->green_mask,
				 image->blue_mask,
				 image->data, width, height,
				 image->bytes_per_line,
				 w->raster.data, w->raster.width,
				 w->raster.height,
				 w->raster.bytes_per_line);
			return;
		    }
		break;
	    case GrayScale:
	    case StaticColor:
	    case StaticGray:
	    case PseudoColor:
		/* For gray scale visuals, this branch is waste with
		   call to Conv24to8. Should have a separate convert
		   to grayscale routine... --msa */
		if (image->bits_per_pixel == 8 && pixelsize >= 24)
		    {
			/*
			  The following code is just thrown together to
			  see that something works. clean out later. --msa */

			unsigned char *tmp = w->raster.data;
			int bps = w->raster.bits_per_sample;

			if (w->raster.width != width ||
			    w->raster.height != height)
			    {
				/* First just scale down by dropping pixels,
				   might consider some bether method here...
				   --msa */
				tmp=(unsigned char *)XtMalloc(width*height*3);
				ImageScaleNtoM(3,
					       tmp, width, height, width * 3,
					       pixelsize / 8,
					       w->raster.data, w->raster.width,
					       w->raster.height,
					       w->raster.bytes_per_line);
			    }
			if (w->raster.color_map)
				XtFree((char *)w->raster.color_map);
			w->raster.color_map =
				(unsigned char *)XtMalloc(3 * 256);
			/* ... This is SLOOOOOW!!!! --msa */
			Conv24to8(tmp,
				  width, height,
				  w->raster.max_colors,
				  (unsigned char*)image->data,
				  w->raster.color_map,
				  w->raster.color_map + 256,
				  w->raster.color_map + 512,
				  w->raster.color_quantize);
			/* Have to fool the XeRasterAllocColor a bit here,
			   obviously this is an indication that the division
			   of work between modules is not well done. Cleaning
			   out needed very seriously... --msa*/
			w->raster.bits_per_sample = 8;
			XeRasterAllocColors(w);
			w->raster.bits_per_sample = bps;
			if (tmp != w->raster.data)
				XtFree((char *)tmp);
			return;
		    }
		break;
	    case DirectColor:
	    default:
		break;
	    }
	Unsupported(w);
    }

void XeRasterLayout(w)
XeRasterWidget w;
    {
	Dimension width, height;
	int format, bitmap_pad, depth, class, size;
	XImage *image;

	if (!fsgamcr_initialized)
	    {
		/* Should do this in class_initialize? --msa */

		GenerateFSGamma();
		fsgamcr_initialized = TRUE;
	    }
	w->raster.done = True;	/* Consider done, whether errors or not */
	if (w->raster.image != NULL)
	    {
		XtFree((char *)w->raster.image->data);
		XFree(w->raster.image);
		w->raster.image = NULL;
	    }
/* Dont Free! colormap may be allocated also due to visual... 
   check this later... --msa
	XeRasterFreeColors(w);
*/
	if (w->raster.visual == NULL)
		w->raster.visual = XeGetVisual((Widget)w);
	class = w->raster.visual->class;

	XeBasicScaling((XeBasicWidget)w,
		       w->raster.width, w->raster.height,
		       &width, &height);
	if (w->raster.data == NULL)
		return;	/* Nothing to layout */
	/*
	** Create New XImage Structure
	*/
	if (w->core.depth == 1 || w->raster.max_colors < 2 ||
	    (w->raster.class == XeRasterClass_BILEVEL &&
	     w->raster.visual->class != PseudoColor))
	    {
		depth = 1;
		bitmap_pad = 32;
		format = XYBitmap;
	    }
	else
	    {
		depth = w->core.depth;
		bitmap_pad = depth > 8 ? 32 : 8;
		format = ZPixmap;
	    }
	w->raster.image = image =
		XCreateImage(XtDisplay(w),
			     w->raster.visual,	/* visual */
			     (int)depth,	/* depth */
			     (int)format,	/* format */
			     0,			/* Offset */
			     (char *)NULL,	/* Bitmap */
			     (unsigned int)width,/* Bitmap width */
			     (unsigned int)height,/* Bitmap height*/
			     (int)bitmap_pad,	/* bitmap_pad */
			     0);		/* Bytes per line */
	if (image == NULL)
		return; /* -- failed in creating the XImage */
#if 0
	printf("XImage: unit=%d, pad=%d, depth=%d, bits=%d\n",
	       image->bitmap_unit,
	       image->bitmap_pad,
	       image->depth,
	       image->bits_per_pixel);
	printf("byte_order=%s, bit_order=%s\n",
	       image->byte_order == MSBFirst ? "MSBFirst" :
	       image->byte_order == LSBFirst ? "LSBFirst" : "Unknown",
	       image->bitmap_bit_order == MSBFirst ? "MSBFirst" :
	       image->bitmap_bit_order == LSBFirst ? "LSBFirst" : "Unknown");
#endif
	size = height * image->bytes_per_line;
	image->data = (char *)XtMalloc(size);
	if (image->data == NULL)
	    {
		XFree(image);
		w->raster.image = NULL;
		return; /* -- failed in allocating image data */
	    }
	switch (w->raster.class)
	    {
	    case XeRasterClass_BILEVEL:
		LayoutBiLevelImage(w, width, height);
		break;
	    case XeRasterClass_GRAYSCALE:
		LayoutGrayScaleImage(w, width, height);
		break;
	    case XeRasterClass_PALETTE:
		LayoutPaletteColorImage(w, width, height);
		break;
	    case XeRasterClass_FULLCOLOR:
		LayoutFullColorImage(w, width, height);
		break;
	    default:
		break;
	    }
    }
